require 'open-uri'
require 'fileutils'
require 'fastimage'
require "io/console"
require 'set'

module Utils
  #画像か？
  def imgexist?(url)
    s = FastImage.size(url)
    s ? s != [0, 0] : s
  end
  
  
  #Web上のバイナリファイルの保存
  # @return [Boolean]
  def getfile(url, filename, max = 0, ua = nil)
    count = 0
    opt = ua ? {'User-Agent' => ua} : {}
    begin
      open(filename, 'wb') do |file|
        URI.open(url, opt) {|data| file.write(data.read)}
      end
      true 
    rescue
      puts "ファイル入出力エラー： " + $!.message.encode("UTF-8")
      count += 1
      return false if count > max  #max回までリトライする
      puts "リトライ： #{count}"
      sleep(1)
      retry
    end
  end
  module_function :imgexist?, :getfile
  
  
  #一文字入力
  # @return [String]
  def key_wait(breakable = true)
    c = nil
    loop {break if (c = STDIN.getch)}
    STDIN.cooked!
    raise Interrupt if breakable and c == "\cC"
    c
  end
  module_function :key_wait
  
  
  #画像ファイルでなければ削除
  # @return [Boolean]
  def delete_non_img(fname)
    if Utils.imgexist?(fname)
      false
    else
      FileUtils.rm(fname)
      true
    end
  end
  module_function :delete_non_img
  
  
  #指定したディレクトリの空フォルダを削除
  def delete_empty_folder(fname = './')
    Dir.chdir(fname)
    Dir.glob("*").each do |name|
      next unless File.directory?(name)
      Dir.rmdir(p name) if Dir.glob("./#{name}/*").empty? rescue next
    end
  end
  module_function :delete_empty_folder
  
  
  #ベルを鳴らす（Ubuntuのみ）
  def bell
    `paplay /usr/share/sounds/freedesktop/stereo/complete.oga`
  end
  module_function :bell
  
  
  #階乗
  # @return [Integer]
  def factorial(n)
    result = 1
    2.upto(n) {|i| result *= i}
    result
  end
  module_function :factorial
  
  
  #a個の中からb個選ぶ順列の数
  # @return [Integer]
  def permutation(a, b)
    [*1..a].last(b).inject(1, &:*)
  end
  
  #aの中からb個選ぶ組み合わせの数
  # @return [Integer]
  def combination(a, b)
    Utils.permutation(a, b) / Utils.factorial(b)
  end
  
  #a個の中から重複を許してb個選ぶ順列の数
  # @return [Integer]
  def repeated_permutation(a, b)
    a ** b
  end
  
  #a個の中から重複を許してb個選ぶ組み合わせの数
  # @return [Integer]
  def repeated_combination(a, b)
    Utils.combination(a + b - 1, b)
  end
  module_function :permutation, :combination, :repeated_permutation, :repeated_combination
  
  
  #アルファベット小文字を使った長さ string_length の文字列を、重複しないようにランダムに num 個生成する。string_length が指定されないときは最短の文字列を使う
  # @return [Array]
  def generate_random_strings(num, string_length = nil)
    table = [*"a".."z"]
    limit = [0, 26, 702, 18278, 475254, 12356630]
    result = []
    generate_string1 = ->(n, l) {
      st = ""
      l.times do
        a, n = n % 26, n / 26
        st = table[a] + st
      end
      st
    }
    generate_string2 = ->(n) {
      idx = limit.find_index {|i| i > n}
      generate_string1.(n - limit[idx - 1], idx)
    }
    
    if string_length and 26 < string_length
      raise "Given length of strings too big."
    end
    
    num_table = Set.new
    if string_length
      n = Utils.repeated_permutation(26, string_length)
      raise "Given length of strings too small." if n < num
      while num_table.size < num
        num_table << rand(n)
      end
      num_table.each {|i| result << generate_string1.(i, string_length)}
    else
      idx = limit.find_index {|i| i >= num}
      raise "Result Array too big." unless idx
      while num_table.size < num
        num_table << rand(limit[idx])
      end
      num_table.each {|i| result << generate_string2.(i)}
    end
    result
  end
  module_function :generate_random_strings
  
  
  #m3u8（HLS）ダウンロード
  def m3u8(url, fname)
    %x(ffmpeg -i #{url} -movflags faststart -c copy -bsf:a aac_adtstoasc #{fname})
  end
  module_function :m3u8
  
  
  #連番ファイル名を sort 順になるように変更する
  def modify_serial_numbers(dir = Dir.pwd)
    Dir.chdir(dir)
    files = Dir.glob("*").select {|fn| File.file?(fn)}
    data1 = []
    data = files.map do |fn|
      m = /^(.*?\D*?)(\d+)(\..+)$/.match(fn)
      if m
        [fn, m[1], m[2].to_i, m[3]]
      else
        data1 << [fn, fn]
        nil
      end
    end.compact
    num_length = data.map {|d| d[2].to_s.length}.max
    data1 += data.map {|d| [d[0], d[1] + ("%0#{num_length}d" % d[2]) + d[3]]}
    
    Dir.mkdir("new_folder")
    data1.each {|be, af| FileUtils.cp(be, "new_folder/#{af}")}
  end
  module_function :modify_serial_numbers
  
  
  #簡易タイマー
  def timer(minutes)
    end_time = Time.now + (minutes * 60).to_i
    
    while Time.now < end_time
      print "\e[2K\e[1Gleft: #{(end_time - Time.now).to_i} sec."
      sleep(1)
    end
    
    Utils.bell
    puts
  end
  module_function :timer
end
